from .graph_task_model import GraphTaskModel
from .node_multiclass_task import NodeMulticlassTask
from .graph_regression_task import GraphRegressionTask
from .graph_binary_classification_task import GraphBinaryClassificationTask
from .qm9_regression import QM9RegressionTask